/* 
 * File   : pathsprite.cpp
 * Version: 2009-10-10
 *
 * Copyright (C) 2010 Jochen Peters
 *
 * dev[at]emirp.net
 * http://www.emirp.net/
 * 
 * This file is part of Approach Control.
 *
 * Approach Control is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Approach Control is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Approach Control. If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <math.h>
#include <SDL.h>
#include <SDL_gfxPrimitives.h>

#include "pathsprite.hpp"


bool ptInRect(int _x, int _y, SDL_Rect* _rect) {
	
	if(_rect->x <= _x  &&  _x < _rect->x+_rect->w  &&  _rect->y <= _y  &&  _y < _rect->y+_rect->h) {
		return true;
	}
	else {
		return false;
	}

}

bool rectInRect(SDL_Rect* _rect1, SDL_Rect* _rect2) {

	for(int i=0; i<2; i++) {
		for(int j=0; j<2; j++) {
			if(ptInRect(_rect1->x + i*_rect1->w, _rect1->y + j*_rect1->h, _rect2)) {
				return true;
			}
		}
	}

	return false;
}


/*****************
 * PathSprite    *
 *****************/


PathSprite::PathSprite(SDL_Surface* _surf, int _posX, int _posY, int _velX, int _velY, float _speed, int _type) {
	
	spritesurf = _surf;
	if(_surf != NULL) {
		pos.w = _surf->w;
		pos.h = _surf->h;
	}
	else {
		pos.w = 0;
		pos.h = 0;
	}

	pos.x = _posX;
	pos.y = _posY;

	centerpos.x = pos.x + pos.w/2;
	centerpos.y = pos.y + pos.h/2;

	if(_speed <= 0.0) {
		speed = 0;
	}
	else if(_speed >= 1.0) {
		speed = MAXSPEED;
	}
	else {
		speed = (int) (MAXSPEED * _speed);
	}
	movcount = MAXSPEED - speed;

	type = _type;

	plength = 0;
	pindex = 0;

	visible = false;
	status = SPRITESTAT_IDLE;

}

void PathSprite::setSurface(SDL_Surface* _srf) {
	
	spritesurf = _srf;
	pos.w = spritesurf->w;
	pos.h = spritesurf->h;

}


bool PathSprite::isVisible() {
	return visible;
}

bool PathSprite::collisionCheck(int _x, int _y) {
	return (visible && ptInRect(_x, _y, &pos));
}

bool PathSprite::collisionCheckInside(PathSprite* _s) {

	if(_s->isVisible()) {
		return (visible && ptInRect(_s->getCPosX(), _s->getCPosY(), &pos));
	}
	else {
		return false;
	}
}

bool PathSprite::collisionCheckOutline(PathSprite* _s) {

	if(_s->isVisible()) {
		return (visible && rectInRect(&pos, _s->getPos()));
	}
	else {
		return false;
	}
}

bool PathSprite::collisionCheckInside(SDL_Rect* _rect) {
	return (visible && ptInRect(centerpos.x, centerpos.y, _rect));
}

bool PathSprite::collisionCheckOutline(SDL_Rect* _rect) {
	return (visible && rectInRect(&pos, _rect));
}

bool PathSprite::checkTarget(Target* _tgt) {

	if(status == SPRITESTAT_ONAPPROACH && type == _tgt->getType() && collisionCheckOutline(&_tgt->pre)) { // PRE-target
		status = SPRITESTAT_FINALAPPROACH;
		return false;
	}

	if(status == SPRITESTAT_FINALAPPROACH && type == _tgt->getType() && collisionCheckInside(&_tgt->pos)) { // real target
		status = SPRITESTAT_LANDED;
		return true;
	}

	return false;

}

bool PathSprite::checkTarget(Target** _tgt, int _length) {
	
	for(int i=0; i<_length; i++) {
		if(checkTarget(_tgt[i])) {
			return true;
		}
	}

	return false;
}

bool PathSprite::checkPath(Target* _tgt) {
	
	if(type == _tgt->getType()) {

		for(int i=(plength-1); i>0; i--) {
			if(_tgt->inTarget(path[i].x, path[i].y)) {
				if(_tgt->inPreTarget(path[i-1].x, path[i-1].y)) {
					return true;
				}
			}
		}
	}

	return false;
}

bool PathSprite::checkPath(Target** _tgt, int _length) {
	
	for(int i=0; i<_length; i++) {
		if(checkPath(_tgt[i])) {
			return true;
		}
	}

	return false;
}


void PathSprite::setPos(int _dstX, int _dstY) {

	pos.x = _dstX;
	pos.y = _dstY;
	centerpos.x = pos.x + pos.w/2;
	centerpos.y = pos.y + pos.h/2;

}


SDL_Rect* PathSprite::getPos() {
	return &pos;
}


void PathSprite::setVelocity(int _velX, int _velY) {

	vel.x = _velX;
	vel.y = _velY;

}


void PathSprite::draw(SDL_Surface* _dstsrf) {
	if(visible) {
		SDL_BlitSurface(spritesurf, NULL, _dstsrf, &pos);
		//SDL_UpdateRects(_dstsrf, 1, &pos);
	}
}

void PathSprite::undraw(SDL_Surface* _dstsrf, Uint32 _bgcolor) {
	SDL_FillRect(_dstsrf, &pos, _bgcolor);
	//SDL_UpdateRects(_dstsrf, 1, &pos);
}

void PathSprite::show(SDL_Surface* _dstsrf) {
	draw(_dstsrf);
	visible = true;
}

void PathSprite::remove(SDL_Surface* _dstsrf, Uint32 _bgcolor) {
	undraw(_dstsrf, _bgcolor);
	visible = false;
	clearPath();
	setPos(0, 0);
	status = SPRITESTAT_IDLE;
}


void PathSprite::calcVelocity() {

	int tmp_x = vel.x;
	int tmp_y = vel.y;
		
	if(plength > 0 && pindex < PMAXLENGTH) { // if path exists
		

		/* Check if next path point is reached */
		while(ptInRect(path[pindex].x, path[pindex].y, &pos)) { // search next point
		//while(vel.x == 0 && vel.y == 0)
			pindex++;
			if(pindex >= plength) { // last point of path reached
				clearPath();
				vel.x = tmp_x;
				vel.y = tmp_y;
				return;
			}

		}

		/* Calc direction vector */
		vel.x = path[pindex].x - centerpos.x;
		vel.y = path[pindex].y - centerpos.y;
		//printf("%d: pt=(%d,%d); pos=(%d,%d); vel=(%d,%d)\n", pindex, path[pindex].x, path[pindex].y, pos.x, pos.y, vel.x, vel.y); // DEBUG

		/* Normalize velocity vector */
		tmp_x = vel.x;
		tmp_y = vel.y;
		if(abs(vel.x) < abs(vel.y)) {
			if(tmp_x == 0) {
				vel.x = 0;
				vel.y = vel.y / abs(tmp_y); // =(-)1
			}
			else {
				if(abs(tmp_y/tmp_x) > 2) {
					vel.x = 0;
				}
				else {
					vel.x = vel.x / abs(tmp_x); // =(-)1
				}
				vel.y = vel.y / abs(tmp_y); // =(-)1
			}
		}
		else if(abs(vel.x) > abs(vel.y)) {
			if(tmp_y == 0) {
				vel.x = vel.x / abs(tmp_x); // =(-)1
				vel.y = 0;
			}
			else {
				vel.x = vel.x / abs(tmp_x); // =(-)1
				if(abs(tmp_x/tmp_y) > 2) {
					vel.y = 0;
				}
				else {
					vel.y = vel.y / abs(tmp_y); // =(-)1
				}
			}
		}
		else {
			vel.x = vel.x / abs(tmp_x); // =(-)1
			vel.y = vel.y / abs(tmp_y); // =(-)1
		}

	}

}


void PathSprite::move(SDL_Surface* _screen, Uint32 _bgcolor) {

	int i;

	if(visible) {
		
		if(movcount > 0) {
			movcount--;
		}
		else {

	
			/* Undraw sprite */
			undraw(_screen, 0);
		
			calcVelocity();
			//printf("%d: pos=(%d,%d); vel=(%d,%d)\n", pindex, pos.x, pos.y, vel.x, vel.y); // DEBUG
		
			/* Move the sprite, bounce at the wall */
			pos.x += vel.x;
			if((pos.x < 0) || (pos.x >= (_screen->w - spritesurf->w))) {
				vel.x = (-1) * vel.x;
				pos.x += vel.x;
				clearPath();
			}
			pos.y += vel.y;
			if ((pos.y < 0) || (pos.y >= (_screen->h - spritesurf->h))) {
				vel.y = (-1) * vel.y;
				pos.y += vel.y;
				clearPath();
			}
		
			/* Calc new centerpos */
			centerpos.x = pos.x + pos.w/2;
			centerpos.y = pos.y + pos.h/2;
		
			/* Blit the sprite onto the screen */
			draw(_screen);
		
			/* Calc new movecount */
			movcount = MAXSPEED - speed;
			if(vel.x != 0 && vel.y != 0) {
				movcount = movcount * ((unsigned int) sqrt(2));
			}

		}

	}

}


int PathSprite::getPosX() {
	return pos.x;
}

int PathSprite::getPosY() {
	return pos.y;
}

int PathSprite::getCPosX() {
	return centerpos.x;
}

int PathSprite::getCPosY() {
	return centerpos.y;
}

int PathSprite::getWidth() {
	return spritesurf->w;
}

int PathSprite::getHeight() {
	return spritesurf->h;
}

int PathSprite::getType() {
	return type;
}


void PathSprite::clearPath() {
	plength = 0;
	pindex = 0;
	if(status == SPRITESTAT_ONPATH) {
		status = SPRITESTAT_FREERUNNING;
	}
}

int PathSprite::addToPath(int _x, int _y) {

	if(plength < PMAXLENGTH) {

		path[plength].x = _x;
		path[plength].y = _y;
		plength++;

		return 1;
	}
	else {
		return 0;
	}

}

int PathSprite::addToPath(Coord* _pt) {
	return addToPath(_pt->x, _pt->y);
}


/* SDL_gfx version */
void PathSprite::drawPathRGBA(SDL_Surface* _dstsrf, Uint8 _r, Uint8 _g, Uint8 _b, Uint8 _a) {

	if(pindex > 0 && plength > 0) { // draw first segment
		lineRGBA(_dstsrf, path[pindex-1].x, path[pindex-1].y, path[pindex].x, path[pindex].y, _r, _g, _b, _a);
	}
	for(int i=(pindex+1); i<plength; i++) { // draw other segements
		lineRGBA(_dstsrf, path[i-1].x, path[i-1].y, path[i].x, path[i].y, _r, _g, _b, _a);
	}

}


/* SDL_Draw version *
void PathSprite::drawPathRGBA(SDL_Surface* _dstsrf, Uint8 _r, Uint8 _g, Uint8 _b, Uint8 _a) {
	
	Uint32 color = SDL_MapRGBA(_dstsrf->format, _r, _g, _b, _a);

	if(pindex > 0 && plength > 0) { // draw first segment
		drawLine(_dstsrf, path[pindex-1].x, path[pindex-1].y, path[pindex].x, path[pindex].y, color);
	}
	for(int i=(pindex+1); i<plength; i++) { // draw other segements
		drawLine(_dstsrf, path[i-1].x, path[i-1].y, path[i].x, path[i].y, color);
	}

}
*/




/*****************
 * Target        *
 *****************/

Target::Target(int _posX, int _posY, int _posW, int _posH, int _preX, int _preY, int _preW, int _preH, int _type) {

	pos.x = _posX;
	pos.y = _posY;
	pos.w = _posW;
	pos.h = _posH;
	pre.x = _preX;
	pre.y = _preY;
	pre.w = _preW;
	pre.h = _preH;

	type = _type;
}

void Target::draw(SDL_Surface* _dstsrf) {
	
	SDL_FillRect(_dstsrf, &pre, SDL_MapRGB(_dstsrf->format, 150, 150, 150));
	//SDL_UpdateRects(_dstsrf, 1, &pre);
	SDL_FillRect(_dstsrf, &pos, SDL_MapRGB(_dstsrf->format, 255, 255, 255));
	//SDL_UpdateRects(_dstsrf, 1, &pos);

}


bool Target::inTarget(int _x, int _y) {
	return ptInRect(_x, _y, &pos);
}

bool Target::inPreTarget(int _x, int _y) {
	return ptInRect(_x, _y, &pre);
}

int Target::getType() {
	return type;
}



/*****************
 * Marker        *
 *****************/

Marker::Marker(PathSprite* _s, unsigned int _startcount) {
	
	sprite = _s;
	sprite->status = SPRITESTAT_WAITING;

	pos.x = _s->getPosX();
	pos.y = _s->getPosY();
	pos.w = _s->getWidth();
	pos.h = _s->getHeight();

	countdown = _startcount;
	
}

void Marker::draw(SDL_Surface* _dstsrf, Uint32 _color) {
	
	SDL_FillRect(_dstsrf, &pos, _color);
	//SDL_UpdateRects(_dstsrf, 1, &pos);

}


bool Marker::startSprite(SDL_Surface* _dstsrf) {
	
	if(countdown > 0) {
		countdown--;
		return false;
	}
	else {
		
		sprite->status = SPRITESTAT_FREERUNNING;
		sprite->show(_dstsrf);
		return true;
	}

}



